package algo

import core._

/**
  * This type represents sequences.
  *
  */

case object SeqTypeKind extends Kind

sealed trait SeqTypeT extends DataTypeT {
  val et: DataTypeT
  val len: ArithTypeT

  override def kind: Kind = SeqTypeKind

  override def superType: TypeT = DataType()

  override def children: Seq[TypeT] = Seq(et, len)
}

final case class SeqType(et: DataTypeT, len: ArithTypeT) extends SeqTypeT {
  override def build(children: Seq[TypeT]): SeqTypeT = SeqType(children(0).asInstanceOf[DataTypeT], children(1).asInstanceOf[ArithTypeT])
}

final case class SeqTypeVar(et: DataTypeT = DataTypeVar(), len: ArithTypeT = ArithTypeVar()) extends SeqTypeT with TypeVarGenT {
  override def buildNewTV(children: Seq[TypeT]): TypeVarGenT = SeqTypeVar(children(0).asInstanceOf[DataTypeT], children(1).asInstanceOf[ArithTypeT])

  override def upperBound: TypeT = SeqType(et, len)
}

/**
  * This type represents an array (i.e. the elements are ordered and can be indexed).
  *
  */

case object ArrayTypeKind extends Kind

sealed trait ArrayTypeT extends SeqTypeT {
  override def kind: Kind = ArrayTypeKind

  override def superType: TypeT = SeqType(et, len)

  override def children: Seq[TypeT] = Seq(et, len)
}

final case class ArrayType(et: DataTypeT, len: ArithTypeT) extends ArrayTypeT {
  override def build(children: Seq[TypeT]): ArrayTypeT = ArrayType(children.head.asInstanceOf[DataTypeT], children(1).asInstanceOf[ArithTypeT])
}

final case class ArrayTypeVar(et: DataTypeT = DataTypeVar(), len: ArithTypeT = ArithTypeVar()) extends ArrayTypeT with TypeVarGenT {
  override def buildNewTV(children: Seq[TypeT]): TypeVarGenT = ArrayTypeVar(children.head.asInstanceOf[DataTypeT], children(1).asInstanceOf[ArithTypeT])

  override def upperBound: TypeT = ArrayType(et, len)
}

case object TupTypeKind extends Kind

sealed trait TupTypeT extends DataTypeT {
  val types: Seq[TypeT]

  override def kind: Kind = TupTypeKind

  override def superType: TypeT = DataType()

  override def children: Seq[TypeT] = types

}

final case class TupType(types: Seq[TypeT]) extends TupTypeT {
  override def build(children: Seq[TypeT]): TupTypeT = TupType(children.asInstanceOf)
}

final case class TupTypeVar(types: Seq[TypeT]) extends TupTypeT with TypeVarGenT {
  override def buildNewTV(children: Seq[TypeT]): TypeVarGenT = TupTypeVar(children.asInstanceOf)

  override def upperBound: DataTypeT = TupType(types)
}

case object AbstractScalarTypeKind extends Kind

trait AbstractScalarTypeT extends DataTypeT {
  override def kind: Kind = AbstractScalarTypeKind

  override def superType: TypeT = DataType()

  override def children: Seq[TypeT] = Seq()
}

final case class AbstractScalarType() extends AbstractScalarTypeT {
  override def build(children: Seq[TypeT]): AbstractScalarTypeT = AbstractScalarType()
}

final case class AbstractScalarTypeVar() extends AbstractScalarTypeT with TypeVarGenT {
  override def buildNewTV(children: Seq[TypeT]): TypeVarGenT = AbstractScalarTypeVar()

  override def upperBound: DataTypeT = AbstractScalarType()
}

case object IntTypeKind extends Kind

trait IntTypeT extends AbstractScalarTypeT {
  val width: ArithTypeT

  override def kind: Kind = IntTypeKind

  override def superType: TypeT = AbstractScalarType()

  override def children: Seq[TypeT] = Seq(width)
}

final case class IntType(width: ArithTypeT) extends IntTypeT {
  override def build(children: Seq[TypeT]): IntTypeT = IntType(children.head.asInstanceOf[ArithTypeT])
}

final case class IntTypeVar(width: ArithTypeT = ArithTypeVar()) extends IntTypeT with TypeVarGenT {
  override def buildNewTV(children: Seq[TypeT]): TypeVarGenT = IntTypeVar(children.head.asInstanceOf[ArithTypeT])

  override def upperBound: TypeT = IntType(width)
}

case object FloatTypeKind extends Kind

trait FloatTypeT extends AbstractScalarTypeT {
  override def kind: Kind = FloatTypeKind

  override def superType: TypeT = AbstractScalarType()
}

final case class FloatType(width: ArithTypeT) extends FloatTypeT {
  override def build(children: Seq[TypeT]): FloatTypeT = FloatType(children.head.asInstanceOf[ArithTypeT])
}

final case class FloatTypeVar(width: ArithTypeT = ArithTypeVar()) extends FloatTypeT with TypeVarGenT {
  override def buildNewTV(children: Seq[TypeT]): TypeVarGenT = FloatTypeVar(children.head.asInstanceOf[ArithTypeT])

  override def upperBound: TypeT = FloatType(width)
}
