package core

sealed trait IR extends TreeNode[IR] {
  val t: TypeT
  def typeChildren: Seq[TypeT]

  override final def build(children: Seq[IR]): IR = build(children, typeChildren)

  def build(children: Seq[IR], typeChildren: Seq[TypeT]): IR

  final def visitTypes(pre: TypeT => Unit, post: TypeT => Unit): Unit = {
    visit(_.typeChildren.foreach(_.visit(pre, post)), ir => ir)
  }

  final def visitAndRebuildTypes(pre: TypeT => TypeT, post: TypeT => TypeT): IR = {
    visitAndRebuild(c => c.build(c.children, c.typeChildren.map(_.visitAndRebuild(pre, post))), ir => ir)
  }

  def hasFreeTypeVar: Boolean = {
    visitTypes({
      case _: TypeVarT => return true
      case _ =>
    }, t => t)
    false
  }

  def toShortString: String = getClass.getSimpleName.split('$').head
}

case class Param(t: TypeT) extends IR {
  override def children: Seq[IR] = Seq()
  override def typeChildren: Seq[TypeT] = Seq(t)
  override def build(children: Seq[IR], typeChildren: Seq[TypeT]): Param = Param(typeChildren.head)
}

final case class BuiltinValue(t: TypeT) extends IR {
  override def children: Seq[IR] = Seq()
  override def typeChildren: Seq[TypeT] = Seq(t)
  override def build(children: Seq[IR], typeChildren: Seq[TypeT]): BuiltinValue = BuiltinValue(typeChildren.head)
}

trait FunctionT extends IR {
  override val t: FunTypeT
}

case class Lambda(param: Param, body: IR, t: FunTypeT) extends FunctionT {
  override def children: Seq[IR] = Seq(param) ++ Seq(body)
  override def typeChildren: Seq[TypeT] = Seq(t)
  override def build(children: Seq[IR], typeChildren: Seq[TypeT]): Lambda = Lambda(children.head.asInstanceOf[Param], children(1), typeChildren.head.asInstanceOf[FunTypeT])
}

object Lambda {
  def apply(param: Param, body: IR): Lambda = Lambda(param, body, FunType(param.t, body.t))

  def apply(params: Seq[Param], body: IR): Lambda = {
    if (params.isEmpty) {
      //TODO
      Lambda(param = null, body)
    } else if (params.size == 1) {
      Lambda(params.last, body)
    } else {
      Lambda(params.last, Lambda(params.init, body))
    }
  }
}

final case class BuiltinFunction(t: FunTypeT) extends FunctionT {
  override def children: Seq[IR] = Seq()
  override def typeChildren: Seq[TypeT] = Seq(t)
  override def build(children: Seq[IR], typeChildren: Seq[TypeT]): BuiltinFunction = BuiltinFunction(typeChildren.head.asInstanceOf[FunTypeT])
}

case class FunctionCall(f: IR, arg: IR, t: TypeT) extends IR {
  override def children: Seq[IR] = Seq(f) ++ Seq(arg)
  override def typeChildren: Seq[TypeT] = Seq(t)
  override def build(children: Seq[IR], typeChildren: Seq[TypeT]): FunctionCall = FunctionCall(children.head, children(1), typeChildren.head)
}

object FunctionCall {
  def apply(f: IR, arg: IR): FunctionCall = FunctionCall(f, arg, AnyTypeVar())

  def apply(f: IR, args: Seq[IR]): FunctionCall = {
    if (args.isEmpty) {
      //TODO
      FunctionCall(f, arg = null)
    } else if (args.size == 1) {
      FunctionCall(f, args.last)
    } else {
      FunctionCall(FunctionCall(f, args.init), args.last)
    }
  }
}

trait TypeFunctionT extends IR {
  override val t: TypeFunTypeT
}

case class TypeLambda(body: IR, t: TypeFunTypeT) extends TypeFunctionT {
  override def children: Seq[IR] = Seq(body)
  override def typeChildren: Seq[TypeT] = Seq(t)
  override def build(children: Seq[IR], typeChildren: Seq[TypeT]): TypeLambda = TypeLambda(children.head, typeChildren.head.asInstanceOf[TypeFunTypeT])
}

object TypeLambda {
  def apply(typeVarParam: TypeVarT, body: IR): TypeLambda = TypeLambda(body, TypeFunType(typeVarParam, body.t))

  def apply(typeVarParams: Seq[TypeVarT], body: IR): TypeLambda = {
    if (typeVarParams.isEmpty) {
      //TODO
      TypeLambda(typeVarParam = null, body)
    } else if (typeVarParams.size == 1) {
      TypeLambda(typeVarParams.last, body)
    } else {
      TypeLambda(typeVarParams.last, TypeLambda(typeVarParams.init, body))
    }
  }
}

case class BuiltinTypeFunction(t: TypeFunTypeT) extends TypeFunctionT {
  override def children: Seq[IR] = Seq()
  override def typeChildren: Seq[TypeT] = Seq(t)
  override def build(children: Seq[IR], typeChildren: Seq[TypeT]): BuiltinTypeFunction = BuiltinTypeFunction(typeChildren.head.asInstanceOf[TypeFunTypeT])
}

case class TypeFunctionCall(tf: IR, arg: TypeT, t: TypeT) extends IR {
  override def children: Seq[IR] = Seq(tf)
  override def typeChildren: Seq[TypeT] = Seq(arg, t)
  override def build(children: Seq[IR], typeChildren: Seq[TypeT]): TypeFunctionCall = TypeFunctionCall(children.head, typeChildren.head, typeChildren(1))
}

object TypeFunctionCall {

  def apply(tf: IR) : IR = {
    TypeFunctionCall(tf, AnyTypeVar())
  }

  def apply(tf: IR, arg: TypeT) : IR = {
    tf match {
      //case TypeLambda(body: IR, t: TypeFunTypeT) => body.replaceTypeVars(Map(t.tv -> arg)) // TODO uncomment smart constructor
      case tFun: TypeFunctionT => new TypeFunctionCall(tFun, arg, tFun.t.outType) // TODO type correct??
      case _ => new TypeFunctionCall(tf, arg, AnyTypeVar())
    }
  }

  def apply(tf: IR, args: Seq[TypeT]): IR = {
    if (args.isEmpty) {
      //TODO
      TypeFunctionCall(tf, arg = null, AnyTypeVar())
    } else if (args.size == 1) {
      TypeFunctionCall(tf, args.last)
    } else {
      TypeFunctionCall(TypeFunctionCall(tf, args.init), args.last)
    }
  }
}

trait Builtin extends IR { parent =>
  override val t: TypeT = AnyTypeVar()
  private[core] val innerIR: IR

  final override def children: Seq[IR] = Seq(innerIR)
  final override def typeChildren: Seq[TypeT] = Seq(t)
  final override def build(childrenP: Seq[IR], typeChildrenP: Seq[TypeT]): Builtin = new Builtin() {
    override val t: TypeT = typeChildrenP.head
    override val innerIR: IR = childrenP.head
    override def toString: String = parent.toShortString
    override def toShortString: String = parent.toShortString
  }
}
