package core

import scala.collection.mutable

object TypeChecker {

  def rebuild(ir: IR): IR = {
    val assignments: Assignments = new Assignments()

    assignments.traverse(ir)

    assignments.solve()

    val typeMap = assignments.getTypeMap
    ir.visitAndRebuildTypes(t => t, {
      case node: TypeVarT => typeMap.getOrElse(node, node)
      case node: TypeT => node
    })
  }

  private class Assignments {

    // lower bounds are further down in the type tree (closer to the leaves), upper bounds are further up in the type tree (closer to the root)
    private case class Bounds(lower: mutable.Set[TypeT], upper: mutable.Set[TypeT]) {
      override def toString: String = "lowerBound" + lower.toString() + "; upperBound" + upper.toString()
    }

    private val map: mutable.Map[TypeVarT, Bounds] = mutable.Map.empty
    private val typeMap: mutable.Map[TypeVarT, TypeT] = mutable.Map.empty

    override def toString: String = "" + map.map(entry => "\n\t" + entry._1 + " -> " + entry._2)

    /**
      * traverses tree and collect assignments (e.g. of types to type variables)
      * also throws exceptions if there are invalid assignments
      *
      * @param ir
      * @return
      */
    def traverse(ir: IR) {
      ir match {
        case Param(_) =>
        case BuiltinFunction(_) =>
        case Lambda(param: Param, body: IR, t: FunTypeT) =>
          traverse(param) // not necessary
          traverse(body)
          add(t.inType, param.t) // not necessary
          add(body.t, t.outType)
        case FunctionCall(f: IR, arg: IR, t: TypeT) =>
          f.t match {
            case FunType(inType: TypeT, outType: TypeT) =>
              // see https://docs.scala-lang.org/tour/variances.html for the order of arguments
              // inType => outType (defined function signature) must be a subtype of arg.t => t (given function signature)
              // ==> arg.t must be a subtype of inType && outType must be a subtype of t
              add(inType, arg.t)
              add(t, outType)
            case tv: TypeVarT =>
              // version long
              // val ftv = FunType(AnyTypeVar(), AnyTypeVar())
              // add(ftv, tv)
              // add(ftv.inType, arg.t)
              // add(t, ftv.outType)
              // version short
              add(FunType(arg.t, t), tv) // well that's a !#&%$ line of code
            case _ => throw new Exception("expected FunType, found " + f.t)
          }
          traverse(arg)
          traverse(f)
        case BuiltinTypeFunction(_) =>
        case TypeLambda(body: IR, t: TypeFunTypeT) =>
          traverse(body)
          add(body.t, t.outType)
        case TypeFunctionCall(tf: IR, arg: TypeT, t: TypeT) =>
          tf.t match {
            case TypeFunType(tv: TypeVarT, outType: TypeT) =>
              add(tv, arg)
              add(t, outType)
            case tv: TypeVarT => ???
              //add(TypeFunType(arg, t), tv)
            case _ => throw new Exception("expected TypeFunType, found " + tf.t)
          }
          traverse(tf)
        case bi: Builtin =>
          add(bi.t, bi.innerIR.t)
          traverse(bi.innerIR)
      }
    }

    def getTypeMap: Map[TypeVarT, TypeT] = {
      typeMap.toMap
    }

    /**
      * extracts assignments (type equations) and checks if both sides are unifiable
      *
      * @param t
      * @param subType
      * @return
      */
    def add(t: TypeT, subType: TypeT): Unit = {
      (t, subType) match {
        // both sides are equal => nothing to do
        case (_, _) if t == subType =>
        case (tv1: TypeVarT, _: TypeVarGenT) if subType.isSubTypeOf(t).getOrElse(true) =>
          addLowerBound(tv1, subType)
          // addUpperBound(tvg2, tv1) // TODO maybe remove because it may create circular dependencies
          addChildren(t, subType) // TODO not needed
        // if there is a typevar => add and check children
        case (tv1: TypeVarT, _) if subType.isSubTypeOf(t).getOrElse(true) =>
          addLowerBound(tv1, subType)
          addChildren(t, subType) // TODO not needed
        case (_, tv2: TypeVarT) if subType.isSubTypeOf(t).getOrElse(true) =>
          addUpperBound(tv2, t)
          addChildren(t, subType) // TODO not needed
        // both sides have equal type (and are not TypeVars!) => check children
        case (_, _) if t.kind == subType.kind =>
          addChildren(t, subType)
        // none of the above => error
        case (_, _) =>
          throw new Exception(t + " not unifiable with subtype " + subType)
      }
    }

    private def addChildren(t: TypeT, subType: TypeT): Unit = {
      if (t.children.size == subType.children.size) {
        if (t.isInstanceOf[FunTypeT] && subType.isInstanceOf[FunTypeT]) {
          add(subType.children.head, t.children.head)
          add(t.children(1), subType.children(1))
        } else {
          t.children.zip(subType.children).foreach(p => add(p._1, p._2))
        }
      }
    }

    private def addLowerBound(tv: TypeVarT, lb: TypeT): Unit = {
      if (map.isDefinedAt(tv)) {
        map(tv).upper.foreach(ub =>
          if (lb.isSubTypeOf(ub) == Option(false)) {
            throw new Exception("cannot add lower bound (" + lb + ") of type, because it does not fit the existing upper bound (" + ub + ")")
          }
        )
        map(tv).lower += lb
      } else {
        tv match {
          //case tvg: TypeVarGenT => map += ((tv, Bounds(mutable.Set(lb), mutable.Set(tvg.upperBound))))
          case _ => map += ((tv, Bounds(mutable.Set(lb), mutable.Set.empty)))
        }
      }
    }

    private def addUpperBound(tv: TypeVarT, ub: TypeT): Unit = {
      if (map.isDefinedAt(tv)) {
        map(tv).lower.foreach(lb =>
          if (lb.isSubTypeOf(ub) == Option(false)) {
            throw new Exception("cannot add upper bound (" + ub + ") of type, because it does not fit the existing lower bound (" + lb + ")")
          }
        )
        map(tv).upper += ub
      } else {
        tv match {
          //case tvg: TypeVarGenT => map += ((tv, Bounds(mutable.Set.empty, mutable.Set(ub, tvg.upperBound))))
          case _ => map += ((tv, Bounds(mutable.Set.empty, mutable.Set(ub))))
        }
      }
    }

    private def mergeAssignments(): Unit = {
      map.foreach(f = elem => {
        val lbs = elem._2.lower
        val uniqueLBPairs = for {
          (x, idxX) <- lbs.zipWithIndex
          (y, idxY) <- lbs.zipWithIndex
          if idxX < idxY
        } yield (x, y)

        uniqueLBPairs.foreach(pair => {
          if (!(pair._1.isInstanceOf[TypeVarT] && pair._2.isInstanceOf[TypeVarT])) { // can only merge if the two elements are both not typevars
            lbs.add(findCommonLowerBound(pair._1, pair._2))
            lbs.remove(pair._1)
            lbs.remove(pair._2)
          }
        })

        val ubs = elem._2.upper
        val uniqueUBPairs = for {
          (x, idxX) <- ubs.zipWithIndex
          (y, idxY) <- ubs.zipWithIndex
          if idxX < idxY
        } yield (x, y)

        uniqueUBPairs.foreach(pair => {
          if (!(pair._1.isInstanceOf[TypeVarT] && pair._2.isInstanceOf[TypeVarT])) { // can only merge if the two elements are both not typevars
            ubs.add(findCommonLowestUpperBound(pair._1, pair._2))
            ubs.remove(pair._1)
            ubs.remove(pair._2)
          }
        })
      })
    }

    /**
      * has side effects!
      *
      * @param t1
      * @param t2
      * @return
      */
    private def findCommonLowerBound(t1: TypeT, t2: TypeT): TypeT = {
      (t1, t2) match {
        case (_, _) if t1.isSubTypeOf(t2) == Option(true) => t2 // TODO maybe there are assignments in the children with typevars to add to the assignments list?????
        case (_, _) if t2.isSubTypeOf(t1) == Option(true) => t1 // TODO maybe there are assignments in the children with typevars to add to the assignments list?????
        case (tv1: TypeVarT, _) =>
          val newTV = tv1.buildNewTV(tv1.children)
          addLowerBound(newTV, t1)
          addLowerBound(newTV, t2)
          newTV
        case (_, tv2: TypeVarT) =>
          val newTV = tv2.buildNewTV(tv2.children)
          addLowerBound(newTV, t1)
          addLowerBound(newTV, t2)
          newTV
        case (ft1: FunTypeT, ft2: FunTypeT) if t1.kind == t2.kind => // special case for funtype
          FunType(findCommonLowestUpperBound(ft1.inType, ft2.inType), findCommonLowerBound(ft1.outType, ft2.outType))
        case (_, _) if t1.kind == t2.kind => // same kinds, create new type of this kind and recursively check children
          t1.build(t1.children.zip(t2.children).map(c => findCommonLowerBound(c._1, c._2)))
        case (_, _) => // neither subtypes, not same kind => find common super type
          findCommonLowerBound(t1.superType, t2.superType)
      }
    }

    /**
      * has side effects!
      *
      * @param t1
      * @param t2
      * @return
      */
    private def findCommonLowestUpperBound(t1: TypeT, t2: TypeT): TypeT = {
      (t1, t2) match {
        case (_, _) if t1.isSubTypeOf(t2) == Option(true) => t1
        case (_, _) if t2.isSubTypeOf(t1) == Option(true) => t2
        case (tv1: TypeVarT, _) =>
          val newTV = tv1.buildNewTV(Seq.empty)
          addUpperBound(newTV, t1)
          addUpperBound(newTV, t2)
          newTV
        case (_, tv2: TypeVarT) =>
          val newTV = tv2.buildNewTV(Seq.empty)
          addUpperBound(newTV, t1)
          addUpperBound(newTV, t2)
          newTV
        case (ft1: FunTypeT, ft2: FunTypeT) if t1.kind == t2.kind => // special case for funtype
          FunType(findCommonLowerBound(ft1.inType, ft2.inType), findCommonLowestUpperBound(ft1.outType, ft2.outType))
        case (_, _) if t1.kind == t2.kind => // same kinds, create new type of this kind and recursively check children
          t1.build(t1.children.zip(t2.children).map(c => findCommonLowestUpperBound(c._1, c._2)))
        case (_, _) => // neither subtypes, not same kind => types do not have anything in common
          throw new Exception("cannot merge upperbounds " + t1 + " and " + t2)
      }
    }

    /**
      * this function should be called, when to types are declared to be equal (during the solving process)
      *
      * @param t1
      * @param t2
      */
    private def addSolvedType(t1: TypeT, t2: TypeT): Unit = {
      (t1, t2) match {
        case (tv1: TypeVarT, tv2: TypeVarT) =>
          _addSolvedType(Seq(tv1), tv1, tv2)
          _addSolvedType(Seq(tv2), tv2, tv1)
        case (tv: TypeVarT, t: TypeT) => _addSolvedType(Seq(tv), tv, t)
        case (t: TypeT, tv: TypeVarT) => _addSolvedType(Seq(tv), tv, t)
        case (_, _) if t1.kind == t2.kind => t1.children.zip(t2.children).foreach(c => addSolvedType(c._1, c._2)) // two types => look for children and solve
        case _ => throw new Exception("Types cannot be equal: " + t1 + " " + t2)
      }
    }

    private def _addSolvedType(visitedTVs: Seq[TypeVarT], tv: TypeVarT, t: TypeT): Unit = {
      typeMap.get(tv) match {
        case Some(tvNew: TypeVarT) =>
          if (!t.isInstanceOf[TypeVarT]) {
            typeMap.put(tv, t) // replace typevar in entry with a specific type
            //typeMap.put(tvNew, t) // not needed will be dealt with recursively (better)
          }
          if (!visitedTVs.contains(tvNew)) { // only go further if there is no cycle dependency (we have not visited this tv's entry yet)
            _addSolvedType(visitedTVs ++ Seq(tvNew), tvNew, t)
          } else { // there is a circle (A=B, B=C, C=A) => break it and insert this new type (C=X)
            typeMap.put(tv, t)
          }
        case Some(tt: TypeT) => // this tv is solved already! solve right hand sides of both equations
          addSolvedType(t, tt)
        case None => typeMap.put(tv, t) // no entry available, add whatever t is
      }
    }


    private def getType(t: TypeT): TypeT = {
      t match {
        case tv: TypeVarT => _getType(Seq(tv), tv)
        case _ => t
      }
    }

    private def _getType(visitedTVs: Seq[TypeVarT], tv: TypeVarT): TypeT = {
      typeMap.get(tv) match {
        case Some(tv_in: TypeVarT) if !visitedTVs.contains(tv_in) => _getType(visitedTVs ++ Seq(tv_in), tv_in)
        case Some(typ: TypeT) => typ
        case None => tv
      }
    }

    /**
      * solves all the assignments
      *
      */
    def solve(): Unit = {
      typeMap.clear()

      mergeAssignments()
      //println("merged assignments: " + toString)

      for ((tv, bounds) <- map) {
        (bounds.lower.size, bounds.upper.size) match {
          case (0, 0) => ??? // this should never happen
          case (0, 1) => addSolvedType(tv, bounds.upper.head) // if there is only an upper bound available, we use this, because we cannot be more specific
          case (1, 0) => addSolvedType(tv, bounds.lower.head) // most specific type (the lower bound)
          case (1, 1) => // map most specific type and assign upper bound to be a subtype (or equal) to lower bound
            addSolvedType(tv, bounds.lower.head)
            addSolvedType(bounds.lower.head, bounds.upper.head) // TODO that can fail? => maybe make this at a later point, where all "simple" typevars are solved already // maybe add extra condition for typevar, if there is one
          case (2, _) => // TODO what if more ....
            addSolvedType(tv, bounds.lower.head)
            addSolvedType(bounds.lower.head, bounds.lower.tail.head) // TODO that can fail?
          case (_, 2) => // TODO what if more ....
            addSolvedType(tv, bounds.lower.head)
            addSolvedType(bounds.lower.head, bounds.lower.tail.head) // TODO that can fail?
          case (_, _) => throw new Exception("assignments must be merged before creating typemap / or implement feature, that can solve multiple upper/lower bounds 'here'")
        }
      }

      var typeMapState = ""

      do {
        // finally replace remaining typevars with upperbounds
        typeMapState = typeMap.toString()
        typeMap.foreach(e => {
          typeMap.put(e._1, e._2.visitAndRebuild(t => t, {
            case node: TypeVarGenT => node.upperBound
            case node: TypeT => node
          }))
        })
      } while (typeMap.toString() != typeMapState) // as long as the type map state is changed, we keep replacing

      do {
        // replace children in upperbounds
        typeMapState = typeMap.toString()
        typeMap.foreach(e => typeMap.put(e._1, e._2.visitAndRebuild(t => t, t => getType(t))))
      } while (typeMap.toString() != typeMapState) // as long as the type map state is changed, we keep replacing
    }
  }

}
