package core

import java.util.concurrent.atomic.AtomicLong

import lift.arithmetic.{ArithExpr, Cst, ExtensibleVar, Range, RangeUnknown, Var}

import scala.language.implicitConversions

object TypeT {
  implicit def aeToATT(ae: ArithExpr): ArithTypeT = ArithType(ae)

  implicit def intToATT(i: Int): ArithTypeT = ArithType(Cst(i))
}

trait Kind

sealed trait TypeT extends TreeNode[TypeT] {
  def kind: Kind

  def superType: TypeT

  final def getSuperKinds: Seq[Kind] = {
    kind match {
      case AnyTypeKind => Seq(kind)
      case _ => Seq(kind) ++ superType.getSuperKinds
    }
  }

  final def getCommonSuperKind(that: TypeT): Kind = {
    val thatSuperKinds = that.getSuperKinds

    for (elem <- this.getSuperKinds) {
      if (thatSuperKinds.contains(elem))
        return elem
    }

    AnyTypeKind
  }

  final def isSubTypeOf(that: TypeT): Option[Boolean] = {
    (this, that) match {
      // Arith (handle these special cases first)
      case (thisTVA: TypeVarArithT, thatTVA: TypeVarArithT) =>
        (thisTVA.range.max - thatTVA.range.max, thisTVA.range.min - thatTVA.range.min) match {
          case (Cst(le), Cst(ge)) => Some(le <= 0 && ge >= 0)
          case _ => None
        }
      case (thisTVA: TypeVarArithT, thatAT: ArithTypeT) =>
        Some(thatAT.ae == thisTVA.range.min && thatAT.ae == thisTVA.range.max)
      case (thisAT: ArithTypeT, thatTVA: TypeVarArithT) =>
        (thisAT.ae - thatTVA.range.min, thisAT.ae - thatTVA.range.max) match {
          case (Cst(ge), Cst(le)) => Some(ge >= 0 && le <= 0) // use ArithExpr.isSmaller(,) if not working
          case _ => None
        }
      case (thisAT: ArithTypeT, thatAT: ArithTypeT) =>
        Some(thisAT.ae == thatAT.ae)
      case (_: TypeVarArithT, _) => None
      // TODO missing? case (_, _ : TypeVarArith) => ???

      // Fun
      // special case: functions are contravariant over its argument type, and covariant over its return type
      case (thisFt: FunTypeT, thatFt: FunTypeT) =>
        (thatFt.inType.isSubTypeOf(thisFt.inType), thisFt.outType.isSubTypeOf(thatFt.outType)) match {
          case (Some(true), Some(true)) => Some(true)
          case (Some(false), _) | (_, Some(false)) => Some(false)
          case (_, _) => None
        }

      // TypeVars
      case (_: TypeVarT, _) if this.kind == AnyTypeKind => None // if this is AnyTypeVar we cannot say anything
      case (_, _: TypeVarT) if that.kind == AnyTypeKind => None // if that is AnyTypeVar we cannot say anything
      case (tv1: TypeVarT, tv2: TypeVarT) if tv1 == tv2 => Some(true)
      case (_, tv: TypeVarGenT) =>
        if (this.getCommonSuperKind(that) == AnyTypeKind)
          Some(false)
        else
          None
      case (tv: TypeVarGenT, _) =>
        if (tv.upperBound.isSubTypeOf(that) == Option(true))
          Some(true)
        else
          None

      // general cases
      case (_, _) if this.kind == that.kind =>
        if (this.children.zip(that.children).forall(p => p._1.isSubTypeOf(p._2).getOrElse(false)))
          Some(true)
        else if (this.children.zip(that.children).forall(p => p._1.isSubTypeOf(p._2).getOrElse(true)))
          None
        else
          Some(false)
      case (_, _) if that.kind == AnyTypeKind => Some(true)
      case (_, _) if this.kind == AnyTypeKind => Some(false)

      // else
      case (_, _) =>
        this.superType.isSubTypeOf(that)
    }
  }

  override def toString: String = getClass.getSimpleName + "[" + children.foldLeft("")((s, t) => s + "*" + t) + "]"

  def toShortString: String = getClass.getSimpleName
}


case object AnyTypeKind extends Kind

sealed trait AnyTypeT extends TypeT {
  override def kind: Kind = AnyTypeKind

  override def superType: TypeT = this

  override def children: Seq[TypeT] = Seq()
}

final case class AnyType() extends AnyTypeT {
  override def build(children: Seq[TypeT]): AnyTypeT = this
}

final case class AnyTypeVar() extends AnyTypeT with TypeVarT {
  override def buildNewTV(children: Seq[TypeT]): TypeVarT = AnyTypeVar()
}

object TypeVarCounter {
  private val cnt = new AtomicLong(0) /* Instance counter */

  def next(): Long = cnt.getAndIncrement()
}

trait TypeVarT extends TypeT {
  // must not override the kind

  override final def build(children: Seq[TypeT]): TypeVarT = this

  def buildNewTV(children: Seq[TypeT]): TypeVarT

  val tvid: Long = TypeVarCounter.next()

  override def equals(that: Any): Boolean =
    that match {
      case thatTV: TypeVarT => this.tvid == thatTV.tvid
      case _ => false
    }

  override def hashCode(): Int = 8 * 79 + tvid.hashCode // need to be the same hash function as in ArithExpr Var class

  override def toString: String = super.toString + "[id=" + tvid + "]"

  override def toShortString: String = getClass.getSimpleName + "(" + tvid + ")"
}

trait TypeVarGenT extends TypeVarT {
  def upperBound: TypeT
}

trait TypeVarArithT extends TypeVarT {
  def range: Range
}


case object FunTypeKind extends Kind

/**
  *
  * @param typeVars : a set of type variables over which this function is generic
  * @param inType   : an input types corresponding to the argument
  * @param outType  : the output type of the function
  */
sealed trait FunTypeT extends AnyTypeT {
  val inType: TypeT
  val outType: TypeT

  override def kind: Kind = FunTypeKind

  override def superType: TypeT = AnyType()

  override def children: Seq[TypeT] = Seq(inType, outType)
}

final case class FunType(inType: TypeT, outType: TypeT) extends FunTypeT {
  override def build(children: Seq[TypeT]): FunTypeT = FunType(children.head, children(1))
}

object FunType {
  def apply(inTypes: Seq[TypeT], outType: TypeT): FunType = {
    if (inTypes.isEmpty) {
      //TODO
      FunType(inType = null, outType)
    } else if (inTypes.size == 1) {
      FunType(inTypes.head, outType)
    } else {
      FunType(inTypes.head, FunType(inTypes.tail, outType))
    }
  }
}

final case class FunTypeVar(inType: TypeT = AnyTypeVar(), outType: TypeT = AnyTypeVar()) extends FunTypeT with TypeVarT {
  override def buildNewTV(children: Seq[TypeT]): FunTypeVar = FunTypeVar(children.head, children(1))
}


case object TypeFunTypeKind extends Kind

sealed trait TypeFunTypeT extends AnyTypeT {
  val tv: TypeVarT
  val outType: TypeT

  override def kind: Kind = TypeFunTypeKind

  override def superType: TypeT = AnyType()

  override def children: Seq[TypeT] = Seq(outType) // ignore tv, because it is bound
}

final case class TypeFunType(tv: TypeVarT, outType: TypeT) extends TypeFunTypeT {
  override def build(children: Seq[TypeT]): TypeT = TypeFunType(tv, children.head) // leave tv as it is, because it is bound and not free
}

object TypeFunType {
  def apply(tvs: Seq[TypeVarT], outType: TypeT): TypeFunType = {
    if (tvs.isEmpty) {
      //TODO
      TypeFunType(tv = null, outType)
    } else if (tvs.size == 1) {
      TypeFunType(tvs.head, outType)
    } else {
      TypeFunType(tvs.head, TypeFunType(tvs.tail, outType))
    }
  }
}

final case class TypeFunTypeVar(tv: TypeVarT = AnyTypeVar(), outType: TypeT = AnyTypeVar()) extends TypeFunTypeT with TypeVarT {
  override def buildNewTV(children: Seq[TypeT]): TypeVarT = TypeFunTypeVar(children.head.asInstanceOf[TypeVarT], children(1))
}

case object ArithTypeKind extends Kind

sealed trait ArithTypeT extends AnyTypeT {
  val ae: ArithExpr

  override def kind: Kind = ArithTypeKind

  override def superType: TypeT = AnyType()

  override def children: Seq[TypeT] = Seq() // TODO look in the ArithExpr for children
}

final case class ArithType(ae: ArithExpr) extends ArithTypeT {
  override def build(children: Seq[TypeT]): ArithTypeT = this // TODO FIXME

  override def map(f: TypeT => TypeT): ArithTypeT = {
    ArithType(ae.visitAndRebuild({
      case at: ArithTypeVar => f(at).asInstanceOf[ArithTypeT].ae
      case x => x
    }))
  }

  override def toString: String = super.toString + "[val=" + ae.toString + "]"

  override def toShortString: String = this.getClass.getSimpleName + "(" + ae.toString + ")"
}

final case class ArithTypeVar(override val range: Range = RangeUnknown, tvFixedId: Option[Long] = None, fixedId: Option[Long] = None) extends ExtensibleVar("", range, fixedId) with ArithTypeT with TypeVarArithT {

  override val tvid: Long = tvFixedId.getOrElse(TypeVarCounter.next())

  val ae: ArithExpr = this

  override def buildNewTV(children: Seq[TypeT]): TypeVarArithT = ArithTypeVar(range = this.range) // TODO FIXME

  override lazy val hashCode: Int = super[TypeVarArithT].hashCode()
  override lazy val toString: String = super[TypeVarArithT].toString()

  override def copy(r: Range) = ArithTypeVar(r)

  override def visitAndRebuild(f: ArithExpr => ArithExpr): ArithExpr = f(ArithTypeVar(range.visitAndRebuild(f), Some(this.tvid), Some(this.id)))
}


//class TypeVar private(range : Range) extends ExtensibleVar("", range) with TypeVarT {
//  override def map(f: (Type) => Type): TypeVar = this
//}

case object DataTypeKind extends Kind

trait DataTypeT extends AnyTypeT {
  override def kind: Kind = DataTypeKind

  override def superType: TypeT = AnyType()
}

final case class DataType() extends DataTypeT {
  override def build(children: Seq[TypeT]): DataTypeT = this
}

final case class DataTypeVar() extends DataTypeT with TypeVarGenT {
  override def buildNewTV(children: Seq[TypeT]): TypeVarT = DataTypeVar()

  override def upperBound: DataTypeT = DataType()
}

