package core.util

import java.io.{File, PrintWriter}
import java.util.concurrent.atomic.AtomicLong

import core.{TypeT, _}

import scala.collection.mutable

object DotPrinter {

  def toFile(ir: IR, fileName: String): Unit = {
    val counter : AtomicLong = new AtomicLong(0)
    val pw = new PrintWriter(new File(fileName))

    pw.write("digraph {\n")
    _visit(ir, pw, counter)
    pw.write("}\n")
    pw.close()
    println("file " + fileName + " written.")
  }

  private def _visit(ir: IR, pw: PrintWriter, counter: AtomicLong) : Long = {
    val myID = counter.getAndIncrement()
    pw.write(s"""node$myID [label=\"${ir.toShortString}\"];\n""")
    ir match {
      case l: Lambda =>
        pw.write(s"""node$myID -> node${_visit(l.param, pw, counter)}[label=\"param\"]\n""")
        pw.write(s"""node$myID -> node${_visit(l.body, pw, counter)}[label=\"body\"]\n""")
      case fc: FunctionCall =>
        pw.write(s"""node$myID -> node${_visit(fc.f, pw, counter)}[label=\"fun\"]\n""")
        pw.write(s"""node$myID -> node${_visit(fc.arg, pw, counter)}[label=\"arg\"]\n""")
      case tl : TypeLambda =>
        pw.write(s"""node$myID -> node${_visit(tl.body, pw, counter)}[label=\"body\"]\n""")
      case tfc: TypeFunctionCall =>
        pw.write(s"""node$myID -> node${_visit(tfc.tf, pw, counter)}[label=\"fun\"]\n""")
        pw.write(s"""node$myID -> node${_visit(tfc.arg, pw, counter)}[label=\"arg\"]\n""")
      case bi: Builtin =>
        pw.write(s"""node$myID -> node${_visit(bi.innerIR, pw, counter)}[label=\"innerIR\"]\n""")
      case _ =>
        ir.children.map(c1 => _visit(c1, pw, counter)).foreach(c2 => pw.write(s"node$myID -> node$c2\n"))
    }
    pw.write(s"""node$myID -> node${_visit(ir.t, pw, counter)} [style=dotted]\n""")// label=\"type\",
    myID
  }

  private def _visit(t: TypeT, pw: PrintWriter, counter: AtomicLong) : Long = {
    val myID = counter.getAndIncrement()

    val nodeStyle = if (t.isInstanceOf[TypeVarT]) ", style=filled" else ""
    pw.write(s"""node$myID [shape=septagon, label=\"${t.toShortString}\"$nodeStyle];\n""")
    t match {
      case ft: FunTypeT =>
        pw.write(s"""node$myID -> node${_visit(ft.inType, pw, counter)} [label=\"in\", style=dotted]\n""")
        pw.write(s"""node$myID -> node${_visit(ft.outType, pw, counter)} [label=\"out\", style=dotted]\n""")
      case tft: TypeFunTypeT =>
        pw.write(s"""node$myID -> node${_visit(tft.tv, pw, counter)} [label=\"tv\", style=dotted]\n""")
        pw.write(s"""node$myID -> node${_visit(tft.outType, pw, counter)} [label=\"out\", style=dotted]\n""")
      case _ =>
        t.children.map(c1 => _visit(c1, pw, counter)).foreach(c2 => pw.write(s"node$myID -> node$c2 [style=dotted]\n"))
    }
    myID
  }


  /**
    * function implementation below
    *
    */

  def toFileFunc(ir: IR, fileName: String, includeTypes: Boolean = true) {
    val pw = new PrintWriter(new File(fileName))
    pw.write("digraph {\n")
    ir.visit(visitFunction(pw, includeTypes))
    pw.write("}\n")
    pw.close()
  }

  private def visitFunction(pw: PrintWriter, includeTypes: Boolean): IR => IR => Unit = {
    val counter: AtomicLong = new AtomicLong(0)
    val idMap = mutable.Map[IR, Long]()
    irPre: IR => {
      val irID = counter.getAndIncrement()
      idMap += (irPre -> irID)
      pw.write(s"""ir$irID [label=\"${irPre.getClass.getSimpleName}\"];\n""")

      if (includeTypes) {
        val typeMap = mutable.Map[TypeT, Long]()
        // function to create tree for types
        irPre.t.visit({
          tPre: TypeT => {
            val typeID = counter.getAndIncrement()
            typeMap += (tPre -> typeID)
            pw.write(s"""type$typeID [shape=septagon, label=\"$tPre\"];\n""")

            tPost: TypeT => {
              tPost.children.map(c1 => typeMap(c1)).foreach(c2 => pw.write(s"type$typeID -> type$c2 [style=dotted]\n"))
            }
          }
        })
        pw.write(s"ir$irID -> type${typeMap.get(irPre.t).get} [style=dotted]\n")
      }

      irPost: IR => {
        irPost match {
          case _: Param =>
          case l: Lambda =>
            pw.write(s"ir$irID -> ir${idMap.get(l.param).get}\n")
            pw.write(s"ir$irID -> ir${idMap.get(l.body).get}\n")
          case _: BuiltinFunction =>
          case fc: FunctionCall =>
            pw.write(s"ir$irID -> ir${idMap.get(fc.f).get}\n")
            pw.write(s"ir$irID -> ir${idMap.get(fc.arg).get}\n")
        }
      }
    }
  }
}

