package vhdl

import core._
import lift.arithmetic.{ArithExpr, StartFromRange}

case object VhdlDataTypekind extends Kind

sealed trait VhdlDataTypeT extends DataTypeT {
  override def kind: Kind = VhdlDataTypekind

  override def superType: TypeT = DataType()

  override def children: Seq[TypeT] = Seq()
}

final case class VhdlDataType() extends VhdlDataTypeT {
  override def build(children: Seq[TypeT]): VhdlDataTypeT = this
}

final case class VhdlDataTypeVar() extends VhdlDataTypeT with TypeVarGenT {
  override def buildNewTV(children: Seq[TypeT]): TypeVarGenT = VhdlDataTypeVar()

  override def upperBound: TypeT = VhdlDataType()
}

/**
  * Types that implements this trait do not have any stream nested inside
  */
case object NoStreamDataTypeKind extends Kind

sealed trait NoStreamDataTypeT extends VhdlDataTypeT {
  override def kind: Kind = NoStreamDataTypeKind

  override def superType: TypeT = VhdlDataType()

  override def children: Seq[TypeT] = Seq()
}

final case class NoStreamDataType() extends NoStreamDataTypeT {
  override def build(children: Seq[TypeT]): NoStreamDataTypeT = this
}

final case class NoStreamDataTypeVar() extends NoStreamDataTypeT with TypeVarGenT {
  override def buildNewTV(children: Seq[TypeT]): TypeVarGenT = NoStreamDataTypeVar()

  override def upperBound: TypeT = NoStreamDataType()
}

// TODO: not sure if tuple type should be able to mix any VHDL type
case object TupleTypeKind extends Kind

sealed trait TupleTypeT extends VhdlDataTypeT {
  val types: Seq[VhdlDataTypeT]

  override def kind: Kind = TupleTypeKind

  override def superType: TypeT = VhdlDataType()

  override def children: Seq[TypeT] = types
}

final case class TupleType(types: Seq[VhdlDataTypeT]) extends TupleTypeT {
  override def build(children: Seq[TypeT]): TupleTypeT = TupleType(children.map(_.asInstanceOf[VhdlDataTypeT]))
}

final case class TupleTypeVar(types: Seq[VhdlDataTypeT]) extends TupleTypeT with TypeVarGenT {
  override def buildNewTV(children: Seq[TypeT]): TypeVarGenT = TupleTypeVar(children.asInstanceOf)

  override def upperBound: TypeT = TupleType(types)
}

case object ArrayTypeKind extends Kind

sealed trait ArrayTypeT extends NoStreamDataTypeT {
  val et: NoStreamDataTypeT
  val len: ArithTypeT

  override def kind: Kind = ArrayTypeKind

  override def superType: TypeT = NoStreamDataType()

  override def children: Seq[TypeT] = Seq(et, len)
}

final case class ArrayType(et: NoStreamDataTypeT, len: ArithTypeT) extends ArrayTypeT {
  if (!ArithExpr.isSmaller(0, len.ae).getOrElse(true)) throw new Exception("invalid length for " + this.getClass)

  override def build(children: Seq[TypeT]): ArrayTypeT = ArrayType(children.head.asInstanceOf[NoStreamDataTypeT], children(1).asInstanceOf[ArithTypeT])
}

final case class ArrayTypeVar(et: NoStreamDataTypeT = NoStreamDataTypeVar(), len: ArithTypeT = ArithTypeVar(StartFromRange(0))) extends ArrayTypeT with TypeVarGenT {
  override def buildNewTV(children: Seq[TypeT]): TypeVarGenT = ArrayTypeVar(children.head.asInstanceOf[NoStreamDataTypeT], children(1).asInstanceOf[ArithTypeT])

  override def upperBound: TypeT = ArrayType(et, len)
}

case object StreamTypeKind extends Kind

sealed trait StreamTypeT extends VhdlDataTypeT {
  val et: VhdlDataTypeT
  val len: ArithTypeT

  override def kind: Kind = StreamTypeKind

  override def superType: TypeT = VhdlDataType()

  override def children: Seq[TypeT] = Seq(et, len)
}

final case class StreamType(et: VhdlDataTypeT, len: ArithTypeT) extends StreamTypeT {
  override def build(children: Seq[TypeT]): StreamTypeT = StreamType(children.head.asInstanceOf[VhdlDataTypeT], children(1).asInstanceOf[ArithTypeT])
}

final case class StreamTypeVar(et: VhdlDataTypeT = VhdlDataTypeVar(), len: ArithTypeT = ArithTypeVar()) extends StreamTypeT with TypeVarGenT {
  override def buildNewTV(children: Seq[TypeT]): TypeVarGenT = StreamTypeVar(children.head.asInstanceOf[VhdlDataTypeT], children(1).asInstanceOf[ArithTypeT])

  override def upperBound: TypeT = StreamType(et, len)
}

case object AbstractScalarTypeKind extends Kind

sealed trait AbstractScalarTypeT extends NoStreamDataTypeT {
  val width: ArithTypeT

  override def kind: Kind = AbstractScalarTypeKind

  override def superType: TypeT = NoStreamDataType()

  override def children: Seq[TypeT] = Seq(width)
}

final case class AbstractScalarType(width: ArithTypeT) extends AbstractScalarTypeT {
  override def build(children: Seq[TypeT]): AbstractScalarTypeT = AbstractScalarType(children.head.asInstanceOf[ArithTypeT])
}

final case class AbstractScalarTypeVar(width: ArithTypeT = ArithTypeVar()) extends AbstractScalarTypeT with TypeVarGenT {
  override def buildNewTV(children: Seq[TypeT]): TypeVarGenT = AbstractScalarTypeVar(children.head.asInstanceOf[ArithTypeT])

  override def upperBound: TypeT = AbstractScalarType(width)
}

case object IntTypeKind extends Kind

sealed trait IntTypeT extends AbstractScalarTypeT {
  val width: ArithTypeT

  override def kind: Kind = IntTypeKind

  override def superType: TypeT = AbstractScalarType(width)

  override def children: Seq[TypeT] = Seq(width)
}

final case class IntType(width: ArithTypeT) extends IntTypeT {
  override def build(children: Seq[TypeT]): IntTypeT = IntType(children.head.asInstanceOf[ArithTypeT])
}

final case class IntTypeVar(width: ArithTypeT = ArithTypeVar()) extends IntTypeT with TypeVarGenT {
  override def buildNewTV(children: Seq[TypeT]): TypeVarGenT = IntTypeVar(children.head.asInstanceOf[ArithTypeT])

  override def upperBound: TypeT = IntType(width)
}

case object FloatTypeKind extends Kind

sealed trait FloatTypeT extends AbstractScalarTypeT {
  val width: ArithTypeT

  override def kind: Kind = FloatTypeKind

  override def superType: TypeT = AbstractScalarType(width)

  override def children: Seq[TypeT] = Seq(width)
}

final case class FloatType(width: ArithTypeT) extends FloatTypeT {
  override def build(children: Seq[TypeT]): FloatTypeT = FloatType(children.head.asInstanceOf[ArithTypeT])
}

final case class FloatTypeVar(width: ArithTypeT = ArithTypeVar()) extends FloatTypeT with TypeVarGenT {
  override def buildNewTV(children: Seq[TypeT]): TypeVarGenT = FloatTypeVar(children.head.asInstanceOf[ArithTypeT])

  override def upperBound: TypeT = FloatType(width)
}

case object FixedTypeKind extends Kind

sealed trait FixedTypeT extends AbstractScalarTypeT {
  val width: ArithTypeT

  override def kind: Kind = FixedTypeKind

  override def superType: TypeT = AbstractScalarType(width)

  override def children: Seq[TypeT] = Seq(width)
}

final case class FixedType(width: ArithTypeT) extends FixedTypeT {
  override def build(children: Seq[TypeT]): FixedTypeT = FixedType(children.head.asInstanceOf[ArithTypeT])
}

final case class FixedTypeVar(width: ArithTypeT = ArithTypeVar()) extends FixedTypeT with TypeVarGenT {
  override def buildNewTV(children: Seq[TypeT]): TypeVarGenT = FixedTypeVar(children.head.asInstanceOf[ArithTypeT])

  override def upperBound: TypeT = FixedType(width)
}
